/*
  Evolve program from the AMBIORE package.

  (C) Copyright 2004, Dick Hwang and University of Washington.

  dhwang@u.washington.edu
  http://www.phrap.org

 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "utility.h"
#include <ctype.h>

#define NUM_NUC 4     /* Size of nucleotide alphabet */
#define GAP 5

#define NUM_ROOT_PARAMETERS 48
#define NUM_TOTAL_ROOT_PARAMETERS 64

#define MAX_NAME 64
#define MAX_LINE 1024
#define MAX_CHILDREN 16
#define NUM_ROOT_BASE 16
#define MAX_BRANCHES 256

#define T 0
#define C 1
#define A 2
#define G 3

#define FLOAT double

#define USAGE "USAGE: evolve [-map mapfile] [-seed seed] [-mutation_types # [filename]] seqfile treefile rate_categories_file root_categories_file\n"


typedef struct NodeS {
  struct NodeS *children[MAX_CHILDREN];
  struct NodeS *parent;
  int n_children;
  double time_to_parent;
  int seq_num;
  int num;
  char* seq;
  char* present;
} Node;

typedef struct {
  char* seq;
  char name[MAX_NAME];
  double time;
} Sequence;

typedef struct {
  FLOAT el[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
} ParamTable;

typedef struct {
  FLOAT el[NUM_NUC][NUM_NUC][NUM_NUC];
} RootTable;

typedef struct {
  signed char rate_category;
  char present; /* 0 = absent, 1 = present, 2 = end node is present */
  signed char direction; /* -1 = reverse, 0 = symmetric, 1 = forward */
} BranchState;

typedef struct {
  int n_seq;
  int num_branches;
  int seq_length;

  int num_parameters[2];
  int num_mutation_parameters[2];
  int num_total_mutation_parameters[2];

  short* root_position;
  BranchState **branch_state; /* indexed by sequence position, then branch */
  int num_rate_categories;
  char* category_type; /* 0 = symmetric, 1 = asymmetric */
  char** category_name; 
  char** category_branch_existence; /* indexed by category, then branch; indicates whether a rate category exists for a particular branch */
  char* root_category;
  int num_root_categories;
  char** root_category_name;

  int* map;
  int map_length;

  ParamTable*** node_param_table;
  char** direction;
  RootTable** node_root_table;

  FLOAT** parameters;
  FLOAT** root_transitions;

  int num_time_categories; /* number of time categories partitioning the branches */
  int* time_category; /* time category of each branch */
  double* totaltime; /* sum of time of each time category */
  int* rate_time_category; /* time category of each rate category */

  Sequence* sequence;
  Node* root;
  RootTable* root_table;
  RootTable* complement_root_table;
  ParamTable** param_table; /* indexed by branch, then category */
  ParamTable** reverse_param_table;
  double branch_ratio[MAX_BRANCHES];
  int branch_length[MAX_BRANCHES];
  double** branch_lengths; /* indexed by mutation type, then branch */

  int num_mutation_types;
  int mutation_types;
  int* mutation_type[2]; /* indexed by symmetry, then mutation parameter */
  int* mutation_type_symmetry; /* indexed by mutation type; 0 = symmetric, 1 = asymmetric, 2 = both */
  int* num_mutation_parameters_per_type[2]; /* indexed by symmetry, then mutation type */
  int** mutation_parameters_per_type[2]; /* indexed by symmetry, then mutation type, then mutation parameter for that type */
  int* num_mutation_types_per_total_parameters[2]; /* indexed by symmetry, then total parameter */
  int** mutation_types_per_total_parameters[2]; /* indexed by symmetry, then total parameter, then mutation type */
  int* num_total_parameters_per_type[2];
  int** total_parameters_per_type[2];

  int trinuc_parameters[2][256]; 
  int complement_trinuc_parameters[256];
  int* param_complement;
  int*** param_enum[2];
  int** param_enum_sum[2];
  int* param_enum_addend[2];
  int num_mutation_set[2];

  int root_parameters[NUM_NUC][NUM_NUC][NUM_NUC];
  int complement_root_parameters[NUM_NUC][NUM_NUC][NUM_NUC];
  int root_parameters_vector[64];
  int complement_root_parameters_vector[64];
  int root_enum[NUM_TOTAL_ROOT_PARAMETERS][3];
  int root_enum_sum[NUM_ROOT_BASE][NUM_NUC];
  int root_enum_addend[NUM_ROOT_PARAMETERS];
} Params;

char NAUC[6] = "TCAGN-";
int complement[4] = {2,3,0,1};



void read_params(int argc, char* argv[], Params* params);
void read_tree(Params* params, char* treefile);
Node* read_node(Params* params, FILE* in, Node* parent);
void count_freqs(Params* params);
void read_sequence(Params* params, char* seqfile);
void read_parameters(Params* params, FLOAT* parameters, int a, char* filename);
void read_root_parameters(Params* params, FLOAT* parameters, char* filename);
void print_tree(Params* params);
void print_node(Params* params, Node* node, int is_root);
void evolve_tree(Params* params);
void evolve_node(Params* params, Node* node, FLOAT* tree, int ancestor_length, char* ancestor);
void generate_sequence(Params* params, int length, char* seq, char* start, char* end, signed char* direction, char* category);
void evolve(Params* params, int length, char* ancestor, char* descendent, double age, FLOAT* tree, signed char* direction, char* category, int node_num);
int get_state(int n, double* vector);
int get_tree_state(int n, FLOAT* tree);
void create_param_table(Params* params);
void print_sequences(Params* params);
void print_node_sequence(Params* params, Node* node);
double random_exponential(double lambda);
void enstate_branches(Params* params);
void instate_node1(Params* params, Node* node);
void instate_node2(Params* params, Node* node);
void read_rate_categories(Params* params, char* filename);
void read_root_categories(Params* params, char* filename);
void read_mapfile(Params* params, char* mapfile);
void make_trinuc_parameters(Params* params, char* mutation_type_file);
void make_root_parameters(Params* params);
void update_param_table(Params* params, int* change);
void read_branch_lengths(Params* params);

int main(int argc, char* argv[]) {
  Params *params;

  (params = (Params*)malloc(sizeof(Params))) || die("Could not allocate memory for params\n");
  
  read_params(argc, argv, params);
  randomize();
  evolve_tree(params);
  print_sequences(params);

  return 0;
}


void print_sequences(Params* params) {
  printf("%d %d\n", params->n_seq, params->seq_length);
  print_node_sequence(params, params->root);
}

void print_node_sequence(Params* params, Node* node) {
  int n;
  int i;

  if(node->n_children)
    for(n = 0; n < node->n_children; n++)
      print_node_sequence(params, node->children[n]);
  else {
    printf("%-10.10s ", params->sequence[node->seq_num].name);
    for(i = 0; i < params->seq_length; i++)
      printf("%c", NAUC[node->seq[i]]);
    printf("\n");
  }
}


void evolve_tree(Params* params) {
  FLOAT *tree;
  char* seq;
  int length;
  int i;
  signed char* direction;
  char* category;
  
  (direction = (signed char*)malloc(sizeof(signed char) * params->seq_length)) || die("Could not allocate memory for direction\n");
  (category = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for category\n");
  
  length = 0;
  for(i = 0; i < params->seq_length; i++)
    if(params->root_position[i] == params->num_branches) {
      direction[length] = params->branch_state[i][params->num_branches].direction;
      category[length] = params->root_category[i];
      length++;
    }
  (seq = (char*)malloc(sizeof(char) * length)) || die("Could not allocate memory for seq\n");

  generate_sequence(params, length, seq, NULL, NULL, direction, category);
  length = 0;
  for(i = 0; i < params->seq_length; i++)
    if(params->root_position[i] == params->num_branches)
      params->root->seq[i] = seq[length++];
  (tree = (FLOAT*)malloc(sizeof(FLOAT) * power(2, ceil(log((double)params->seq_length * (NUM_NUC - 1)) / log(2.0)) + 1))) || die("Could not allocate memory for tree\n");

  evolve_node(params, params->root, tree, length,  seq);

  free(seq);
  free(direction);
  free(category);
  free(tree);
}

void evolve_node(Params* params, Node* node, FLOAT* tree, int ancestor_length, char* ancestor) {
  int n;
  int i, length, node_length, next;
  char* seq1;
  char* seq2;
  signed char* direction;
  char* category;

  if(!node->children)
    return;

  (seq1 = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for seq\n");

  (seq2 = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for seq\n");

  (direction = (signed char*)malloc(sizeof(signed char) * params->seq_length)) || die("Could not allocate memory for direction\n");
  (category = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for root_category\n");

  

  for(n = 0; n < node->n_children; n++) {
    fprintf(stderr, ".");
    length = node_length = 0;
    for(i = 0; i < params->seq_length; i++)
      if(params->branch_state[i][node->num].present) {
	if(params->branch_state[i][node->children[n]->num].present == 1) {
	  direction[length] = params->branch_state[i][node->children[n]->num].direction;
	  category[length] = params->branch_state[i][node->children[n]->num].rate_category;
	  seq1[length++] = ancestor[node_length++];
	}
	else
	  node_length++;
      }

    evolve(params, length, seq1, seq2, 1.0, tree, direction, category, node->children[n]->num);

    next = length = node_length = 0;
    for(i = 0; i < params->seq_length; i++)
      if(params->branch_state[i][node->children[n]->num].present) {
	category[node_length] = params->root_category[i];
	direction[node_length] = params->branch_state[i][node->children[n]->num].direction;
	if(params->branch_state[i][node->children[n]->num].present == 1) {
	  if(next != node_length) {
	    generate_sequence(params, node_length - next, seq1 + next, seq1 + next - 2, seq2 + length, direction + next, category + next);
	  }
	  seq1[node_length++] = seq2[length++];
	  next = node_length;
	}
	else {
	  node_length++;
	}
      }
    length = 0;
    for(i = 0; i < params->seq_length; i++)
      if(params->branch_state[i][node->children[n]->num].present)
	node->children[n]->seq[i] = seq1[length++];
      else
	node->children[n]->seq[i] = GAP;

    evolve_node(params, node->children[n], tree, node_length, seq1);
  }

  free(seq1);
  free(category);
  free(direction);
  free(seq2);
}

void generate_sequence(Params* params, int length, char* seq, char *start, char* end, signed char *direction, char* category) {
  int i;
  char cap[2];
  RootTable* root_table;
  

  do {
    if(start) {
      seq[0] = start[0];
      if(length > 1)
	seq[1] = start[1];
    }
    else {
      seq[0] = random_int(0, 3);
      if(length > 1)
	seq[1] = random_int(0, 3);
    }
    for(i = 2; i < length; i++) {
      if(category[i] < 0 || category[i] >= params->num_root_categories)
      fprintf(stderr, "%d,", category[i]);
      root_table = (direction[i] >= 0 ? params->root_table : params->complement_root_table) + category[i];
      seq[i] = get_state(NUM_NUC, root_table->el[seq[i-2]][seq[i-1]]);
    }

    root_table = (direction[i] >= 0 ? params->root_table : params->complement_root_table) + category[i-1];
    cap[0] = get_state(NUM_NUC, root_table->el[seq[i-2]][seq[i-1]]);
    root_table = (direction[i+1] >= 0 ? params->root_table : params->complement_root_table) + category[i-1];
    cap[1] = get_state(NUM_NUC, root_table->el[seq[i-1]][cap[0]]);
      
  } while(0 && !(end && (length <= 2 ||  cap[0] == end[0] && cap[1] == end[1]) || !end && (length <= 2 || cap[0] == seq[0] && cap[1] == seq[1])));
}

void evolve(Params* params, int length, char* ancestor, char* descendent, double age, FLOAT* tree, signed char *direction, char* category, int node_num) {
  double sum;
  int i, j, k, to, change, tree_size, bottom, high, low;

  bottom = power(2, ceil(log((double)length * (NUM_NUC - 1)) / log(2.0)));
  tree_size = bottom * 2;

  memcpy(descendent, ancestor, sizeof(char) * length);

  for(k = 0; k < NUM_NUC - 1; k++) {
    to = k + ((k >= descendent[0]) ? 1 : 0);
    tree[bottom + k] = direction[0] >= 0 ? params->param_table[node_num][category[0]].el[descendent[params->seq_length - 1]][descendent[0]][descendent[1]][to] : params->reverse_param_table[node_num][category[0]].el[descendent[params->seq_length - 1]][descendent[0]][descendent[1]][to];
  }
  for(j = 1; j < length - 1; j++)
    for(k = 0; k < NUM_NUC - 1; k++) {
      to = k + ((k >= descendent[j]) ? 1 : 0);
      tree[bottom + j * (NUM_NUC - 1) + k] = direction[j] >= 0 ? params->param_table[node_num][category[j]].el[descendent[j-1]][descendent[j]][descendent[j+1]][to] : params->reverse_param_table[node_num][category[j]].el[descendent[j-1]][descendent[j]][descendent[j+1]][to];
    }
  for(k = 0; k < NUM_NUC - 1; k++) {
    to = k + ((k >= descendent[j]) ? 1 : 0);
    tree[bottom + j * (NUM_NUC - 1) + k] = direction[j] >= 0 ? params->param_table[node_num][category[j]].el[descendent[j-1]][descendent[j]][0][to] : params->reverse_param_table[node_num][category[j]].el[descendent[j-1]][descendent[j]][0][to];
  }
  for(j = bottom + (length + 1) * (NUM_NUC - 1) - 1; j < tree_size; j++)
    tree[j] = 0.0;
  for(i = bottom - 1; i >= 1; i--)
    tree[i] = tree[2*i] + tree[2*i+1];
  change = -1;
  while(1) {
    if(change >= 0) {
      low = (change - 1 >= 0) ? change - 1 : 0;
      high = (change + 1 <= length - 1) ? change + 1 : (length - 1);
      for(j = low; j <= high; j++)
	for(k = 0; k < NUM_NUC - 1; k++) {
	  to = k + ((k >= descendent[j]) ? 1 : 0);
	  tree[bottom + j * (NUM_NUC - 1) + k] = direction[j] >= 0 ? params->param_table[node_num][category[j]].el[descendent[j ? (j-1) : (length - 1)]][descendent[j]][descendent[(j < length - 1) ? (j+1) : 0]][to] : params->reverse_param_table[node_num][category[j]].el[descendent[j ? (j-1) : (length - 1)]][descendent[j]][descendent[(j < length - 1) ? (j+1) : 0]][to];
	}
      low = low * (NUM_NUC - 1) + bottom ;
      high = (high + 1) * (NUM_NUC - 1) - 1 + bottom;
      while(1) {
	low /= 2;
	high /= 2;
	if(low == 0)
	  break;
	for(i = low; i <= high; i++)
	  tree[i] = tree[2*i] + tree[2*i+1];
      }
    }
    age -= random_exponential(tree[1]);
    if(age <= 0.0)
      break;
    change = get_tree_state(tree_size, tree);
    descendent[change / (NUM_NUC - 1)] = change % (NUM_NUC - 1) + ((change % (NUM_NUC - 1) >= descendent[change / (NUM_NUC - 1)]) ? 1 : 0);
    change /= (NUM_NUC - 1);
  }
}

double random_exponential(double lambda) {
  double rand;

  do
    rand = random_double(0.0, 1.0);
  while(rand == 0.0);
  return -log(rand) / lambda;
}

int get_tree_state(int tree_size, FLOAT* tree) {
  double r;
  int i;

  r = random_double(0.0, tree[1]);
  for(i = 2; i < tree_size; i *= 2)
    if(r > tree[i]) {
      r -= tree[i];
      i++;
    }
  return i / 2 - tree_size / 2;
}

int get_state(int n, double* vector) {
  double r;
  int high, low, test;

  r = random_double(0.0, vector[n-1]);

  low = 0;
  high = n - 1;
  while (1) {
    if(low == high)
      return low;
    test = (high + low) / 2;
    if(r < vector[test])
      high = test;
    else
      low = test + 1;
  }
}



void read_params(int argc, char* argv[], Params* params) {
  int i, ip = 0;
  int seed = 0;
  char* mapfile = NULL;
  char* mutation_type_file;

  params->num_mutation_types = 1;

  for(i = 1; i < argc && argv[i][0] == '-'; i++) {
    if(!strcmp(argv[i] + 1, "map"))
      mapfile = argv[++i];
    else if(!strcmp(argv[i] + 1, "seed"))
      seed = atoi(argv[++i]);
    else if(!strcmp(argv[i] + 1, "mutation_types")) {
      params->num_mutation_types = atoi(argv[++i]);
      if(params->num_mutation_types > 1)
	mutation_type_file = argv[++i];
    }
    else
      die(USAGE);
  }

  argc == i + 4 || die(USAGE);
  fprintf(stderr, "Reading parameters...\n");

  read_sequence(params, argv[i++]);
  read_tree(params, argv[i++]);
  read_mapfile(params, mapfile);

  make_trinuc_parameters(params, mutation_type_file);
  make_root_parameters(params);

  read_branch_lengths(params);

  fprintf(stderr, "Enstating branches...\n");
  enstate_branches(params);
  fprintf(stderr, "Reading rate categories...\n");
  read_rate_categories(params, argv[i++]);
  fprintf(stderr, "Reading root categories...\n");
  read_root_categories(params, argv[i++]);

    
  if(seed)
    srandom(seed);
  else
    randomize();
}

void read_branch_lengths(Params* params) {
  int t, b;
  FILE* in;
  char line[MAX_LINE];
  char inname[MAX_LINE];
  char** mutation_type_name;

  (params->branch_lengths = (double**)malloc(sizeof(double*) * params->num_mutation_types)) || die("Could not allocate memory for branch_length\n");
  for(t = 0; t < params->num_mutation_types; t++)
    (params->branch_lengths[t] = (double*)malloc(sizeof(double) * params->num_branches)) || die("Could not allocate memory for params->branch_length[%d]\n", t);
  

  for(t = 0; t < params->num_mutation_types; t++) {
    sprintf(inname, "mcmc.mutationtype_%.2d.branch.list", t);
    (in = fopen(inname, "r")) || die("Could not open file: %s\n", inname);
    
    for(b = 0; b < params->num_branches; b++) {
      fgets(line, MAX_LINE, in) || die("Error reading file: %s\n", inname);
      sscanf(line, "%lg", params->branch_lengths[t] + b) == 1 || die("Error reading line: %s\n", line);
    }
  }
}

void enstate_branches(Params* params) {
  int i, j, count;

  (params->root_position = (short*)malloc(sizeof(short) * params->seq_length)) || die("Could not allocate memory for params->root_position\n");
  (params->branch_state = (BranchState**)malloc(sizeof(BranchState*) * params->seq_length)) || die("Could not allocate memory for params->branch_state\n");

  for(i = 0; i < params->seq_length; i++) {
    (params->branch_state[i] = (BranchState*)malloc(sizeof(BranchState) * (params->num_branches + 1))) || die("Could not allocate memory for params->branch_state[%d]\n", i);
    params->root_position[i] = -1;
  }

  instate_node1(params, params->root);
  instate_node2(params, params->root);
}

void instate_node1(Params* params, Node* node) {
  int i, n;
  char* seq;

  if(!node->n_children) {
    seq = params->sequence[node->seq_num].seq;
    for(i = 0; i < params->seq_length; i++)
      params->branch_state[i][node->num].present = (*seq++ <= NUM_NUC);
  }
  else {
    for(n = 0; n < node->n_children; n++)
      instate_node1(params, node->children[n]);
    for(i = 0; i < params->seq_length; i++) {
      params->branch_state[i][node->num].present = 0;
      for(n = 0; n < node->n_children; n++)
	if(params->branch_state[i][node->children[n]->num].present)
	  params->branch_state[i][node->num].present++;
    }
  }
}
  
void instate_node2(Params* params, Node* node) {
  int i, n;

  for(i = 0; i < params->seq_length; i++) {
    if(params->branch_state[i][node->num].present > 1 || params->branch_state[i][node->num].present && node->parent && params->branch_state[i][node->parent->num].present || params->branch_state[i][node->num].present && !node->n_children) {
      if(!node->parent || !params->branch_state[i][node->parent->num].present) {
	params->branch_state[i][node->num].present = 2;
	params->root_position[i] = node->num;
      }
      else {
	params->branch_state[i][node->num].present = 1;
      }
    }
    else
      params->branch_state[i][node->num].present = 0;
  }
  for(n = 0; n < node->n_children; n++)
    instate_node2(params, node->children[n]);
}

void update_root_table(Params* params) {
  int i, j, k, l, m, c, s, t;
  double prob[2][NUM_NUC][NUM_NUC];
  double sum;



  (params->root_table = (RootTable*)malloc(sizeof(RootTable) * params->num_root_categories)) || die("Could not allocate memory for params->root_table\n");
  (params->complement_root_table = (RootTable*)malloc(sizeof(RootTable) * params->num_root_categories)) || die("Could not allocate memory for params->complement_root_table\n");
  
  for(c = 0; c < params->num_root_categories; c++) {
    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++) {
	params->root_table[c].el[i][j][NUM_NUC - 1] = 1.0;
	for(k = 0; k < NUM_NUC - 1; k++) {
	  params->root_table[c].el[i][j][k] = params->root_transitions[c][params->root_parameters[i][j][k]];
	  params->root_table[c].el[i][j][NUM_NUC - 1] -= params->root_table[c].el[i][j][k];
	}
      }
    
    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++)
	prob[0][i][j] = 1.0 / 16.0;
    for(m = 0; m < 100; m++) {
      s = m % 2;
      t = (m+1) % 2;
      sum = 0.0;
      for(j = 0; j < NUM_NUC; j++)
	for(k = 0; k < NUM_NUC; k++) {
	  prob[t][j][k] = 0;
	  for(i = 0; i < NUM_NUC; i++)
	    prob[t][j][k] += prob[s][i][j] * params->root_table[c].el[i][j][k];
	  sum += prob[t][j][k];
	}
      for(j = 0; j < NUM_NUC; j++)
	for(k = 0; k < NUM_NUC; k++)
	  prob[t][j][k] /= sum;
    }
    
    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++)
	for(k = 0; k < NUM_NUC; k++)
	  params->complement_root_table[c].el[i][j][k] = params->root_table[c].el[complement[k]][complement[j]][complement[i]] * prob[1][complement[k]][complement[j]] / prob[1][complement[j]][complement[i]];

    for(i = 0; i < NUM_NUC; i++)
      for(j = 0; j < NUM_NUC; j++)
	for(k = 1; k < NUM_NUC; k++) {
	  params->root_table[c].el[i][j][k] += params->root_table[c].el[i][j][k-1];
	  params->complement_root_table[c].el[i][j][k] += params->complement_root_table[c].el[i][j][k-1];
	}
  }
}

void read_mapfile(Params* params, char* mapfile) {
  int i;
  FILE *in;
  int length;

  if(!mapfile) {
    (params->map = (int*)malloc(sizeof(int) * params->seq_length)) || die("Could not allocate memory for params->map\n");
    params->map_length = params->seq_length;
    for(i = 0; i < params->seq_length; i++)
      params->map[i] = i;
  }
  else {
    (in = fopen(mapfile, "r")) || die("Could not open file: %s\n", mapfile);
    fscanf(in, "%d", &params->map_length) == 1 || die("Error reading length in mapfile: %s\n", mapfile);
    (params->map = (int*)malloc(sizeof(int) * params->map_length)) || die("Could not allocate memory for params->map\n");
    for(i = 0; i < params->map_length; i++) {
      fscanf(in, "%d", params->map + i) == 1 || die("Error reading value %d in mapfile: %s\n", i, mapfile);
      params->map[i] >= -1 && params->map[i] < params->seq_length || die("Map value is out of range: %d\n", params->map[i]);
    }
  }
}

void make_trinuc_parameters(Params* params, char* mutation_type_file) {
  int i, j, k, l, a;
  int count;
  int tmp;
  FILE* in;
  char line[MAX_LINE];
  char c[8];
  int value;
  

  for(a = 0; a < 2; a++) {
    if(a) {
      params->num_mutation_parameters[a] = 192;
      params->num_total_mutation_parameters[a] = 256;
      params->num_mutation_set[a] = 1;
      (params->param_complement = (int*)malloc(sizeof(int) * params->num_total_mutation_parameters[a])) || die("Could not allocate memory for params->param_complement\n");
    }
    else {
      params->num_mutation_parameters[a] = 96;
      params->num_total_mutation_parameters[a] = 128;
      params->num_mutation_set[a] = 2;
    }


    (params->mutation_type[a] = (int*)malloc(sizeof(int) * params->num_mutation_parameters[a])) || die("Could not allocate memory for params->mutation_type\n");
    (params->num_mutation_parameters_per_type[a] = (int*)malloc(sizeof(int) * params->num_mutation_types)) || die("Could not allocate memory for params->num_mutation_parameters_per_type\n");
    (params->num_total_parameters_per_type[a] = (int*)malloc(sizeof(int) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->num_total_parameters_per_type\n");
    (params->num_mutation_types_per_total_parameters[a] = (int*)malloc(sizeof(int) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->num_mutation_types_per_total_parameters[%d]\n", a);
    (params->mutation_types_per_total_parameters[a] = (int**)malloc(sizeof(int*) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->mutation_types_per_total_parameters[%d]\n", a);
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++)
      (params->mutation_types_per_total_parameters[a][i] = (int*)malloc(sizeof(int) * 2)) || die("Could not allocate memory for params->mutation_types_per_total_parameters[%d][%d]\n", a, i);

    (params->param_enum[a] = (int***)malloc(sizeof(int**) * params->num_total_mutation_parameters[a])) || die("Could not allocate memory for params->param_enum[%d]\n", a);
    for(i = 0; i < params->num_total_mutation_parameters[a]; i++) {
      (params->param_enum[a][i] = (int**)malloc(sizeof(int*) * params->num_mutation_set[a])) || die("Could not allocate memory for params->param_enum[%d]\n", i);
      for(j = 0; j < params->num_mutation_set[a]; j++)
	(params->param_enum[a][i][j] = (int*)malloc(sizeof(int) * 4)) || die("Could not allocate memory for params->param_enum[%d][%d][%d]\n", a, i, j);
    }
    (params->param_enum_sum[a] = (int**)malloc(sizeof(int*) * (params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]))) || die("Could not allocate memory for params->param_enum_sum[%d]\n", a);
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++)
      (params->param_enum_sum[a][i] = (int*)malloc(sizeof(int) * 3)) || die("Could not allocate memory for params->param_enum_sum[%d][%d]\n", a, i);
    (params->param_enum_addend[a] = (int*)malloc(sizeof(int) * params->num_mutation_parameters[a])) || die("Could not allocate memory for params->param_enum_addend[%d]\n", a);
    
    
    for(i = 0; i < 256; i++)
      params->trinuc_parameters[a][i] = -1;
    
    tmp = params->num_mutation_set[a] - 1;

    count = 0;
    for(k = 0; k < NUM_NUC; k++)
      for(l = 0; l < NUM_NUC; l++)
	for(i = 0; i < NUM_NUC; i++)
	  for(j = 0; j < NUM_NUC; j++)
	    if(l != k && params->trinuc_parameters[a][64*i + 16*k + 4*j + l] == -1) {
	      params->trinuc_parameters[a][64*i + 16*k + 4*j + l] = count;
	      params->param_enum[a][count][0][0] = i;
	      params->param_enum[a][count][0][1] = k;
	      params->param_enum[a][count][0][2] = j;
	      params->param_enum[a][count][0][3] = l;
	      params->mutation_type[a][count] = 0;
	      if(a) {
		params->complement_trinuc_parameters[64 * complement[j] + 16 * complement[k] + 4 * complement[i] + complement[l]] = count;
		params->param_complement[count] = count + 1;
		params->param_complement[count+1] = count;
		count++;
		params->complement_trinuc_parameters[64*i + 16*k + 4*j + l] = count;
	      }
	      params->trinuc_parameters[a][64 * complement[j] + 16 * complement[k] + 4 * complement[i] + complement[l]] = count;
	      params->param_enum[a][count][tmp][0] = complement[j];
	      params->param_enum[a][count][tmp][1] = complement[k];
	      params->param_enum[a][count][tmp][2] = complement[i];
	      params->param_enum[a][count][tmp][3] = complement[l];
	      count++;
	    }

    for(i = 0; i < NUM_NUC; i++)
      for(k = 0; k < NUM_NUC; k++)
	for(j = 0; j < NUM_NUC; j++)
	  if(params->trinuc_parameters[a][64*i + 17*k + 4*j] == -1) {
	    params->trinuc_parameters[a][64*i + 17*k + 4*j] = count;
	    params->param_enum[a][count][0][0] = i; 
	    params->param_enum[a][count][0][1] = k;
	    params->param_enum[a][count][0][2] = j;
	    params->param_enum[a][count][0][3] = k;
	    if(a) {
	      params->complement_trinuc_parameters[64*complement[j] + 17*complement[k] + 4*complement[i]] = count;
	      params->param_complement[count] = count + 1;
	      params->param_complement[count+1] = count;
	      count++;
	      params->complement_trinuc_parameters[64*i + 17*k + 4*j] = count;
	    }
	    params->trinuc_parameters[a][64*complement[j] + 17*complement[k] + 4*complement[i]] = count;
	    params->param_enum[a][count][tmp][0] = complement[j];
	    params->param_enum[a][count][tmp][1] = complement[k];
	    params->param_enum[a][count][tmp][2] = complement[i];
	    params->param_enum[a][count][tmp][3] = complement[k];
	    count++;
	  }
    count == params->num_total_mutation_parameters[a] || die("miscounted parameters: %d\n", count);
    
    for(i = params->num_mutation_parameters[a]; i < params->num_total_mutation_parameters[a]; i++)
      for(j = 0; j < 3; j++) {
	params->param_enum_sum[a][i - params->num_mutation_parameters[a]][j] = params->trinuc_parameters[a][64*params->param_enum[a][i][0][0] + 16*params->param_enum[a][i][0][1] +4*params->param_enum[a][i][0][2] + ((j < params->param_enum[a][i][0][3]) ? j : (j+1))];
      }
    for(i = 0; i < params->num_mutation_parameters[a]; i++) {
      params->param_enum_addend[a][i] = params->trinuc_parameters[a][64 * params->param_enum[a][i][0][0] + 16*params->param_enum[a][i][0][1] + 4*params->param_enum[a][i][0][2] + params->param_enum[a][i][0][1]];
    }

    if(params->num_mutation_types == 1)
      for(i = 0; i < params->num_mutation_parameters[a]; i++)
	params->mutation_type[a][i] = 0;
    else if(!a) {
      (in = fopen(mutation_type_file, "r")) || die("Could not open file: %s\n", mutation_type_file);
      for(i = 0; i < params->num_mutation_parameters[a]; i++) {
	fgets(line, MAX_LINE, in) || die("Error in %s reading line: %s\n", mutation_type_file, line);
	sscanf(line, "%c%c%c->%c = %c%c%c->%c = %d", c, c+1, c+2, c+3, c+4, c+5, c+6, c+7, &value) == 9 || die("Error in %s reading line: %s\n", mutation_type_file, line);
	if(NAUC[params->param_enum[a][i][0][0]] != c[0] || NAUC[params->param_enum[a][i][0][1]] != c[1] || NAUC[params->param_enum[a][i][0][2]] != c[2] || NAUC[params->param_enum[a][i][0][3]] != c[3] || NAUC[params->param_enum[a][i][1][0]] != c[4] || NAUC[params->param_enum[a][i][1][1]] != c[5] || NAUC[params->param_enum[a][i][1][2]] != c[6] || NAUC[params->param_enum[a][i][1][3]] != c[7])
	  die("Incorrect ordering of mutations in file %s on line: %s\n", mutation_type_file, line);
	value < params->num_mutation_types || die("%d mutation types specified, but read %d in file %s\n", params->num_mutation_types, value, mutation_type_file);
	params->mutation_type[a][i] = value;
      }
    }
    else {
      for(i = 0; i < params->num_mutation_parameters[a]; i++) {
	if(fgets(line, MAX_LINE, in) && sscanf(line, "%c%c%c->%c = %d", c, c+1, c+2, c+3, &value) == 5) {
	  if(NAUC[params->param_enum[a][i][0][0]] != c[0] || NAUC[params->param_enum[a][i][0][1]] != c[1] || NAUC[params->param_enum[a][i][0][2]] != c[2] || NAUC[params->param_enum[a][i][0][3]] != c[3])
	    die("Incorrect ordering of mutations in file %s on line: %s\n", mutation_type_file, line);
	  value < params->num_mutation_types || die("%d mutation types specified, but read %d in file %s\n", params->num_mutation_types, value, mutation_type_file);
	  params->mutation_type[a][i] = value;
	}
	else
	  params->mutation_type[a][i] = params->mutation_type[0][i/2];
      }
      fclose(in);
    }

    for(i = 0; i < params->num_mutation_types; i++)
      params->num_mutation_parameters_per_type[a][i] = params->num_total_parameters_per_type[a][i] = 0;
    for(i = 0; i < params->num_mutation_parameters[a]; i++)
      params->num_mutation_parameters_per_type[a][params->mutation_type[a][i]]++;
    
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
      params->num_mutation_types_per_total_parameters[a][i] = 0;
      for(j = 0; j < 3; j++) {
	k = params->mutation_type[a][params->param_enum_sum[a][i][j]];
	for(l = 0; l < params->num_mutation_types_per_total_parameters[a][i]; l++)
	  if(params->mutation_types_per_total_parameters[a][i][l] == k)
	    break;
	if(l == params->num_mutation_types_per_total_parameters[a][i]) {
	  params->mutation_types_per_total_parameters[a][i][params->num_mutation_types_per_total_parameters[a][i]++] = k;
	  params->num_total_parameters_per_type[a][k]++;
	}
      }
    }

    (params->mutation_parameters_per_type[a] = (int**)malloc(sizeof(int*) * params->num_mutation_types)) || die("Could not allocate memory for params->mutation_parameters_per_type\n");
    (params->total_parameters_per_type[a] = (int**)malloc(sizeof(int*) * params->num_mutation_types)) || die("Could not allocate memory for params->total_parameters_per_type[%d]\n", a);
    for(i = 0; i < params->num_mutation_types; i++) {
      if(params->num_mutation_parameters_per_type[a][i])
	(params->mutation_parameters_per_type[a][i] = (int*)malloc(sizeof(int) * params->num_mutation_parameters_per_type[a][i])) || die("Could not allocate memory for params->mutation_parameters_per_type[%d]\n", i);
      if(params->num_total_parameters_per_type[a][i])
	(params->total_parameters_per_type[a][i] = (int*)malloc(sizeof(int) * params->num_total_parameters_per_type[a][i])) || die("Could not allocate memory for params->total_parameters_per_type[%d][%d]\n", a, i);
    }
    for(i = 0; i < params->num_mutation_types; i++)
      params->num_mutation_parameters_per_type[a][i] = params->num_total_parameters_per_type[a][i] = 0;
    for(i = 0; i < params->num_mutation_parameters[a]; i++)
      params->mutation_parameters_per_type[a][params->mutation_type[a][i]][params->num_mutation_parameters_per_type[a][params->mutation_type[a][i]]++] = i;  
    for(i = 0; i < params->num_total_mutation_parameters[a] - params->num_mutation_parameters[a]; i++) {
      for(j = 0; j < params->num_mutation_types_per_total_parameters[a][i]; j++)
	params->total_parameters_per_type[a][params->mutation_types_per_total_parameters[a][i][j]][params->num_total_parameters_per_type[a][params->mutation_types_per_total_parameters[a][i][j]]++] = i;
    }
  }  

  (params->mutation_type_symmetry = (int*)malloc(sizeof(int) * params->num_mutation_types)) || die("Could not allocate memory for params->mutation_type_symmetry\n");
  
  for(i = 0; i < params->num_mutation_types; i++) {
    params->mutation_type_symmetry[i] = 0;
    if(params->num_mutation_parameters_per_type[1][i]) {
      params->mutation_type_symmetry[i] = 1;
      if(params->num_mutation_parameters_per_type[0][i])
	params->mutation_type_symmetry[i] = 2;
    }
    fprintf(stderr, "mutation_type_symmetry[%d] = %d\n", i, params->mutation_type_symmetry[i]);
  }
}

  

void make_root_parameters(Params* params) {
  int i, j, k;
  int count = 0;
  
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
     for(k = 0; k < NUM_NUC; k++)
	params->root_parameters[i][j][k] = -1;
  
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC - 1; k++)
	if(params->root_parameters[i][j][k] == -1) {
	  params->root_parameters[i][j][k] = count;
	  params->root_parameters_vector[16 * i + 4 * j + k] = count;
	  params->complement_root_parameters[complement[k]][complement[j]][complement[i]] = count;
	  params->complement_root_parameters_vector[16 * complement[k] + 4 * complement[j] + complement[i]] = count;
	  params->root_enum[count][0] = i;
	  params->root_enum[count][1] = j;
	  params->root_enum[count][2] = k;
	  count++;
	}
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      if(params->root_parameters[i][j][NUM_NUC - 1] == -1) {
	params->root_parameters[i][j][NUM_NUC - 1] = count;
	params->root_parameters_vector[16 * i + 4 * j + 3] = count;
	params->complement_root_parameters[complement[NUM_NUC - 1]][complement[j]][complement[i]] = count;
	params->complement_root_parameters_vector[16 * complement[3] + 4 * complement[j] + complement[i]] = count;
	params->root_enum[count][0] = i;
	params->root_enum[count][1] = j;
	params->root_enum[count][2] = NUM_NUC - 1;
	count++;
      }
  count == NUM_TOTAL_ROOT_PARAMETERS || die("miscounted root parameters: %d\n", count);

  for(i = NUM_ROOT_PARAMETERS; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
    for(j = 0; j < 3; j++)
      params->root_enum_sum[i-NUM_ROOT_PARAMETERS][j] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][j];
  for(i = 0; i < NUM_ROOT_PARAMETERS; i++)
    params->root_enum_addend[i] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][NUM_NUC - 1];
}

void read_rate_categories(Params* params, char* filename) {
  FILE* in;
  int c, i, j, k, l, m;
  char name[MAX_NAME];
  int start, end, num_regions, num_branches, *region_start, *region_end, *branch_start, *branch_end, category;
  int direction;
  char parameters_file[MAX_NAME];
  int tmp, last;

  for(i = 0; i < params->seq_length; i++) {
    for(j = 0; j < params->num_branches; j++) {
      params->branch_state[i][j].rate_category = 0;
      params->branch_state[i][j].direction = 0;
    }
  }
  (in = fopen(filename, "r")) || die("Could not open rate categories file: %s\n", filename);

  (params->time_category = (int*)malloc(sizeof(int) * (params->num_branches + 1))) || die("Could not allocate memory for params->time_category\n");
  for(i = 0; i < params->num_branches; i++)
    params->time_category[i] = 0;
  fscanf(in, "%d", &params->num_time_categories) == 1 || die("Error reading number of time categories in file: %s\n", filename);
  (params->totaltime = (double*)malloc(sizeof(double) * params->num_time_categories)) || die("Could not allocate memory for params->num_time_categories\n");
  for(c = 0; c < params->num_time_categories; c++) {
    fscanf(in, "%d", &num_branches) == 1 || die("Error reading number of branch ranges in time category %d\n", c);
    if(num_branches)
      for(i = 0; i < num_branches; i++) {
	fscanf(in, "%d-%d", &start, &end) == 2 || die("Error reading branch range %d in time category %d\n", i, c);
	start <= end && start >= 0 && end <= params->num_branches || die("Branch range %d to %d is not valid in file: %s\n", start, end, filename);
	for(j = start; j <= end; j++)
	  params->time_category[j] = c;
      }
    else
      for(i = 0; i < params->num_branches; i++)
	params->time_category[i] = c;
    
    params->totaltime[c] = 0.0;
  }

  for(i = 0; i < params->num_branches; i++)
    params->totaltime[params->time_category[i]] += params->branch_ratio[i] * params->branch_length[i];
  for(c = 0; c < params->num_time_categories; c++)
    fprintf(stderr, "totaltime[%d] = %g\n", c, params->totaltime[c]);

  fscanf(in, "%d", &params->num_rate_categories) == 1 || die("Error reading number of rate categories in file: %s\n", filename);
  (params->category_name = (char**)malloc(sizeof(char*) * params->num_rate_categories)) || die("Could not allocate memory for params->category_name\n");
  (params->category_type = (char*)malloc(sizeof(char) * params->num_rate_categories)) || die("Could not allocate memory for params->category_type\n");
  (params->parameters = (FLOAT**)malloc(sizeof(FLOAT*) * params->num_rate_categories)) || die("Could not allocate memory for params->parameters\n");
  (params->category_branch_existence = (char**)malloc(sizeof(char*) * params->num_rate_categories)) || die("Could not allocate memory for params->category_branch_existence\n");
  (params->rate_time_category = (int*)malloc(sizeof(int) * params->num_rate_categories)) || die("Could not allocate memory for params->rate_time_category\n");

  for(c = 0; c < params->num_rate_categories; c++) {
    fscanf(in, "%s %d %s %d", name, &tmp, parameters_file, params->rate_time_category + c) <= 4 || die("Error reading category in file: %s\n", filename);
    params->category_type[c] = tmp;
    params->category_name[c] = strdup(name);
    params->category_type[c] == 0 || params->category_type[c] == 1 || die("Invalid category type: %d\n", params->category_type);
    params->rate_time_category[c] >= 0 && params->rate_time_category[c] < params->num_time_categories || die("Time category %d of rate category is out of range\n", params->rate_time_category[c]);
    (params->parameters[c] = (FLOAT*)malloc(sizeof(FLOAT) * params->num_mutation_parameters[params->category_type[c]])) || die("Could not allocate memory for params->parameters[%d]\n", c);
    (params->category_branch_existence[c] = (char*)malloc(sizeof(char) * params->num_branches)) || die("Could not allocate memory for params->category_branch_existence[%d]\n", c);
    for(i = 0; i < params->num_branches; i++)
      params->category_branch_existence[c][i] = 0;
    read_parameters(params, params->parameters[c], params->category_type[c], parameters_file);
  }

  /*
  for(c = 0; c < params->num_rate_categories; c++)
    for(i = 0; i < params->num_mutation_parameters[params->category_type[c]]; i++)
      fprintf(stderr, "parameter[c][i] = %g\n", params->parameters[c][i]);
  */
  (branch_start = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_start\n");
  (branch_end = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_end\n");
  while(1) {
    tmp = fscanf(in, "%s %d %d", name, &direction, &num_branches);
    if(tmp == EOF)
      break;
    else if(tmp < 3)
      die("Error reading listing in file: %s\n", filename);
    for(i = 0; i < params->num_rate_categories; i++)
      if(!strcmp(name, params->category_name[i])) {
	category = i;
	break;
      }
    i < params->num_rate_categories || die("Unknown category name in rate categories file: %s\n", name);
    if(num_branches)
      for(i = 0; i < num_branches; i++) {
	fscanf(in, "%d-%d", branch_start + i, branch_end + i) == 2 || die("Missing branch in branch list in file: %s\n", filename);
	branch_start[i] <= branch_end[i] && branch_start[i] >= 0 && branch_end[i] <= params->num_branches || die("Invalid branch region: %d to %d\n", branch_start[i], branch_end[i]);
      }
    else {
      branch_start[0] = 0;
      branch_end[0] = params->num_branches;
      num_branches = 1;
    }

    fscanf(in, "%d", &num_regions) == 1 || die("Missing number of regions in file: %s\n", filename);
    if(num_regions)
      for(i = 0; i < num_regions; i++) {
	fscanf(in, "%d-%d", &start, &end) == 2 || die("Missing region in region list in file: %s\n", filename);
	start > 0 && end <= params->map_length || die("Invalid region: %d to %d\n", start, end);
	last = -1;
	for(j = start - 1; j < end; j++)
	  if(params->map[j] >= 0) {
	    for(m = ((last >= 0) ? last : params->map[j]); m <= params->map[j]; m++)
	      for(k = 0; k < num_branches; k++)
		for(l = branch_start[k]; l <= branch_end[k]; l++) {
		  params->branch_state[m][l].rate_category = category;
		  params->branch_state[m][l].direction = direction;
		}
	    last = m;
	  }
      }
    else {
      for(i = 0; i < params->seq_length; i++)
	for(k = 0; k < num_branches; k++)
	  for(l = branch_start[k]; l <= branch_end[k]; l++) {
	    params->branch_state[i][l].rate_category = category;
	    params->branch_state[i][l].direction = direction;
	  }
    }
  }
    
  free(branch_start);
  free(branch_end);
  fclose(in);

  for(i = 0; i < params->seq_length; i++)
    for(j = 0; j <= params->num_branches; j++)
      if(params->branch_state[i][j].present == 1)
	params->category_branch_existence[params->branch_state[i][j].rate_category][j] = 1;
  update_param_table(params, NULL);
}


void update_param_table(Params* params, int* change) {
  int i, j, k, l,  b, c, p, a;
  FLOAT tmp;

  (params->param_table = (ParamTable**)malloc(sizeof(ParamTable*) * params->num_branches)) || die("Could not allocate memory for params->branch_param_table\n");
  (params->reverse_param_table = (ParamTable**)malloc(sizeof(ParamTable*) * params->num_branches)) || die("Could not allocate memory for params->reverse_branch_param_table\n");
  for(i = 0; i < params->num_branches; i++) {
    (params->param_table[i] = (ParamTable*)malloc(sizeof(ParamTable) * params->num_rate_categories)) || die("Could not allocate memory for params->branch_param_table\n");
    (params->reverse_param_table[i] = (ParamTable*)malloc(sizeof(ParamTable) * params->num_rate_categories)) || die("Could not allocate memory for params->reverse_branch_param_table\n");
  }

  for(c = 0; c < params->num_rate_categories; c++) {
    a = params->category_type[c];
    for(b = 0; b < params->num_branches; b++)
      if(params->category_branch_existence[c][b])
	for(i = 0; i < NUM_NUC; i++)
	  for(k = 0; k < NUM_NUC; k++) 
	    for(j = 0; j < NUM_NUC; j++) {
	      params->param_table[b][c].el[i][k][j][k] = 0.0;
	      for(l = 0; l < NUM_NUC; l++)
		if(l != k) {
		  p = params->trinuc_parameters[params->category_type[c]][64 * i + 16 * k + 4 * j + l];
		  tmp = params->param_table[b][c].el[i][k][j][l] = params->branch_lengths[params->mutation_type[a][p]][b] * params->parameters[c][p];
		  params->param_table[b][c].el[i][k][j][k] -= tmp;
		  if(params->category_type[c])
		    params->reverse_param_table[b][c].el[complement[j]][complement[k]][complement[i]][complement[l]] = tmp;
		}
	      if(params->category_type[c])
		params->reverse_param_table[b][c].el[complement[j]][complement[k]][complement[i]][complement[k]] = params->param_table[b][c].el[i][k][j][k];
	      /*
		for(l = 1; l < NUM_NUC; l++) {
		params->param_table[c].el[i][k][j][l] += params->param_table[c].el[i][k][j][l-1];
		if(params->category_type[c])
		params->reverse_param_table[c].el[i][k][j][l] += params->reverse_param_table[c].el[i][k][j][l-1];
		}
	      */
	    }
  }
}

void read_parameters(Params* params, FLOAT* parameters, int a, char* filename) {
  int i;
  FILE* in;
  double parameter;
  char line[MAX_LINE];
  char nuc[8];
  char alpha_table[256];

  (in = fopen(filename, "r")) || die("Could not open parameters file: %s\n", filename);
  for(i = 0; i < NUM_NUC; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < params->num_mutation_parameters[a]; i++) {
    fgets(line, MAX_LINE, in) || die("Truncated parameters file at line %d\n", i);
    if(a)
      sscanf(line, "%c%c%c->%c = %lf", nuc, nuc+1, nuc+2, nuc+3, &parameter) == 5 || 
	sscanf(line, "%c%c%c->%c %lf", nuc, nuc+1, nuc+2, nuc+3, &parameter) == 5 || 
	die("Error reading line: %s", line);
    else
      sscanf(line, "%c%c%c->%c = %c%c%c->%c = %lf", nuc, nuc+1, nuc+2, nuc+3, nuc+4, nuc+5, nuc+6, nuc+7, &parameter) == 9 || 
	sscanf(line, "%c%c%c->%c = %c%c%c->%c %lf", nuc, nuc+1, nuc+2, nuc+3, nuc+4, nuc+5, nuc+6, nuc+7, &parameter) == 9 || 
	die("Error reading line: %s", line);
    parameters[i] = parameter;
  }

  fclose(in);
}

void read_root_categories(Params* params, char* filename) {
  FILE* in;
  int c, i, j, k, l, m;
  char name[MAX_NAME];
  int start, end, num_regions, num_branches, *branch_start, *branch_end;
  char root_file[MAX_NAME];
  int tmp, category, last;

  (params->root_category = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->root_category\n");
  for(i = 0; i < params->seq_length; i++)
    params->root_category[i] = 0;
  (in = fopen(filename, "r")) || die("could not open root categories file: %s\n", filename);
  fscanf(in, "%d", &params->num_root_categories) == 1 || die("Error reading number of root categories in file: %s\n", filename);

  (params->root_category_name = (char**)malloc(sizeof(char*) * params->num_root_categories)) || die("Could not allocate memory for params->root_category_name\n");
  (params->root_transitions = (FLOAT**)malloc(sizeof(FLOAT*) * params->num_root_categories)) || die("Could not allocate memory for params->root_transitions\n");

  for(c = 0; c < params->num_root_categories; c++) {
    fscanf(in, "%s %s", name, root_file) == 2 || die("Error reading root category in file: %s\n", filename);
    params->root_category_name[c] = strdup(name);
    (params->root_transitions[c] = (FLOAT*)malloc(sizeof(FLOAT) * NUM_ROOT_PARAMETERS)) || die("Could not allocate memory for params->root_transitions[%d]\n", i);
    read_root_parameters(params, params->root_transitions[c], root_file);
  }
  update_root_table(params);

  (branch_start = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_start\n");
  (branch_end = (int*)malloc(sizeof(int) * params->num_branches)) || die("Could not allocate memory for branch_end\n");
  while(1) {
    tmp = fscanf(in, "%s %d", name, &num_branches);
    if(tmp == EOF)
      break;
    else if(tmp < 2)
      die("Error reading listing in file: %s\n", filename);
    for(i = 0; i < params->num_root_categories; i++) 
      if(!strcmp(name, params->root_category_name[i])) {
	category = i;
	break;
      }
    i < params->num_root_categories || die("Unknown category name in root categories file: %s\n", name);
    if(num_branches)
      for(i = 0; i < num_branches; i++) {
	fscanf(in, "%d-%d", branch_start + i, branch_end + i) == 2 || die("Missing branch in branch list in file: %s\n", filename);
	branch_start[i] <= branch_end[i] && branch_start[i] >= 0 && branch_end[i] <= params->num_branches || die("Invalid branch region: %d to %d\n", branch_start[i], branch_end[i]);
      }
    else {
      branch_start[0] = 0;
      branch_end[0] = params->num_branches;
      num_branches = 1;
    }

    fscanf(in, "%d", &num_regions) == 1 || die("Could not read number of regions in file: %s\n", filename);
    if(num_regions)
      for(i = 0; i < num_regions; i++) {
	fscanf(in, "%d-%d", &start, &end) == 2 || die("Could not read region in list in file: %s\n", filename);
	start > 0 && end <= params->map_length || die("Invalid region: %d to %d\n", start, end);
	last = -1;
	for(j = start - 1; j < end; j++)
	  if(params->map[j] >= 0) {
	    for(m = ((last >= 0) ? last : params->map[j]); m <= params->map[j]; m++)
	      for(k = 0; k < num_branches; k++)
		if(params->root_position[m] >= branch_start[k] && params->root_position[m] <= branch_end[k])
		  params->root_category[m] = category;
	    last = m;
	  }
      }
    else 
      for(j = 0; j < params->seq_length; j++)
	for(k = 0; k < num_branches; k++)
	  if(params->root_position[j] >= branch_start[k] && params->root_position[j] <= branch_end[k])
	    params->root_category[j] = category;
  }
  free(branch_start);
  free(branch_end);
  fclose(in);
}


void read_root_parameters(Params* params, FLOAT* parameters, char* filename) {
  FILE* in;
  double parameter;
  char line[MAX_LINE];
  char nuc[6];
  char alpha_table[256];
  int i, j, k;
  double root[NUM_NUC][NUM_NUC][NUM_NUC];

  (in = fopen(filename, "r")) || die("Could not open root parameters file: %s\n", filename);
  for(i = 0; i < NUM_NUC; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++) {
    fgets(line, MAX_LINE, in) || die("Truncated root parameters file at line %d\n", i);
    sscanf(line, "%c%c%c = %lf", nuc, nuc+1, nuc+2, &parameter) == 4 || die("Error reading line: %s", line);
    root[alpha_table[nuc[0]]][alpha_table[nuc[1]]][alpha_table[nuc[2]]] = parameter;
  }
  fclose(in);

  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++) {
      root[i][j][NUM_NUC - 1] = 1.0;
      for(k = 0; k < NUM_NUC - 1; k++) {
	root[i][j][NUM_NUC - 1] -= root[i][j][k];
	parameters[params->root_parameters[i][j][k]] = root[i][j][k];
      }
    }

}

void read_sequence(Params* params, char* seqfile) {
  FILE* in;
  int i, j, k, l, m, ch = EOF;
  signed char alpha_table[256];

  (in = fopen(seqfile, "r")) || die("Could not open seqfile: %s\n", seqfile);
  fscanf(in, "%d%d\n", &params->n_seq, &params->seq_length) == 2 || die("Error in seqfile: %s\n", seqfile);
  (params->sequence = (Sequence*)malloc(sizeof(Sequence) * params->n_seq)) || die("Could not allocate memory for params->sequence\n");

  for(i = 0; i < 256; i++)
    alpha_table[i] = -1;
  for(i = 0; i < 6; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < params->n_seq; i++) {
    fscanf(in, "%s.63", params->sequence[i].name) == 1 || die("Error in seqfile\n");
    (params->sequence[i].seq = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->sequence[%d].seq\n", i);
    for(j = 0; j < params->seq_length;) {
      if((ch = getc(in)) != EOF && alpha_table[ch] >= 0) {
	params->sequence[i].seq[j] = alpha_table[ch];
	j++;
      }
      else if(ch == '\n' || ch == EOF)
	die("Early sequence termination in sequence: %s, %d expected, %d found\n", params->sequence[i].name, params->seq_length, j + 1);
    }
    for(; ch != '\n' && ch != EOF; ch = getc(in));
  }
  
  fclose(in);
}

void print_tree(Params* params) {
  if(params->root)
    print_node(params, params->root, 1);
}

void print_node(Params* params, Node* node, int is_root) {
  int n;

  if(node->n_children) {
    printf("(");
    for(n = 0; n < node->n_children; n++) {
      if(n)
	printf(",");
      print_node(params, node->children[n], 0);
    }
    printf(")");
  }
  else
    printf(params->sequence[node->seq_num].name);
  if(!is_root)
    printf(":%g", node->time_to_parent);
  else
    printf(";\n");
}

void read_tree(Params* params, char* treefile) {
  FILE* in;

  (in = fopen(treefile, "r")) || die("Could not open treefile: %s\n", treefile);
  params->num_branches = 0;
  params->root = read_node(params, in, NULL);
  fclose(in);
  params->root->num = params->num_branches;
}


Node* read_node(Params* params, FILE* in, Node* parent) {
  Node* node;
  int i, ch;
  char name[MAX_NAME];
  double branch_length;

  (node = (Node*)malloc(sizeof(Node))) || die("Could not allocate memory for node\n");
  (node->seq = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for node->seq\n");
  node->parent = parent;
  node->n_children = 0;

  while((ch = getc(in)) == '\n');
  if(ch == '(') {
    while(1) {
      while(isspace(ch = getc(in)));
      if(ch == ')')
	break;
      node->n_children < MAX_CHILDREN || die("Node has too many children\n");
      if(node->n_children)
	ch == ',' || die("Missing comma in treefile\n");
      else
	ungetc(ch, in);
      node->children[node->n_children++] = read_node(params, in, node);
      node->seq_num = -1;
    }
  }
  else {
    ungetc(ch, in);
    fscanf(in, "%[^\n:].63", name) == 1 || die("Missing name in treefile\n");
    for(i = 0; i < params->n_seq; i++)
      if(!strcmp(params->sequence[i].name, name))
	break;
    if(i == params->n_seq)
      die("Name %s in tree file but not found in sequence file\n", name);
    node->seq_num = i;
  }
  if(parent) {
    while(isspace(ch = getc(in)));
    ch == ':' || die("Missing colon in treefile after %s\n", name);
    while(isspace(ch = getc(in)));
    ungetc(ch, in);
    fscanf(in, "%lf", &node->time_to_parent) == 1 || die("Missing time to parent\n"); 
    node->time_to_parent = 1.0;
    node->num = params->num_branches++;
  }
  else {
    node->time_to_parent = 1;
    while(isspace(ch = getc(in)));
    ch == ';' || die("Missing semicolon at end of treefile: %c\n", ch);
  }
  return node;
}
