#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "utility.h"

#define USAGE "USAGE: parsimony seqfile treefile\n"

#define NUM_NUC 4     /* Size of nucleotide alphabet */
#define NUM_PARAMETERS 48

#define MAX_NAME 64
#define MAX_CHILDREN 16
#define BIG_SHORT 128

typedef struct {
  char el[NUM_NUC];
} NucVec;

typedef struct {
  short el[NUM_NUC];
} ShortNucVec;

typedef struct NodeS {
  struct NodeS *children[MAX_CHILDREN];
  int n_children;
  double time_to_parent;
  ShortNucVec* num_changes;
  NucVec* point;
  int seq;
  int branch_num;
  char* pars_seq;
} Node;

typedef struct {
  char* seq;
  char name[MAX_NAME];
  double time;
} Sequence;

typedef struct {
  int n_seq;
  int seq_length;
  int num_branches;
  Sequence* sequence;
  char* valid_sequence;
  Node* root;
} Params;

typedef struct {
  int trinuc[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  double length;
} Branch;

typedef struct {
  int mononuc[NUM_NUC][NUM_NUC];
  int monoback[NUM_NUC];
  int symmononuc[NUM_NUC][NUM_NUC];
  int symmonoback[NUM_NUC];
  int dinuc[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int diback[NUM_NUC][NUM_NUC];
  int symdinuc[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int symdiback[NUM_NUC][NUM_NUC];
  int trinuc[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int triback[NUM_NUC][NUM_NUC][NUM_NUC];
  double symdinucrate[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int samples;
  double time;
  Branch* branch;
} Changes;

char NAUC[7] = "TCAGN-";
char NCUC[7] = "AGTCN-";
int complement[4] = {2,3,0,1};

int param_enum[NUM_PARAMETERS][2][4] = {
  {{0,0,0,1},{2,2,3,2}}, {{0,0,0,2},{2,2,0,2}}, {{0,0,0,3},{2,2,1,2}},
  {{0,0,1,0},{2,2,2,3}}, {{0,0,2,0},{2,2,2,0}}, {{0,0,3,0},{2,2,2,1}},

  {{0,1,0,0},{3,2,2,2}}, {{0,1,0,2},{3,2,0,2}}, {{0,1,0,3},{3,2,1,2}},
  {{0,1,1,1},{3,2,3,3}}, {{0,1,2,1},{3,2,3,0}}, {{0,1,3,1},{3,2,3,1}},

  {{0,2,0,0},{0,2,2,2}}, {{0,2,0,1},{0,2,3,2}}, {{0,2,0,3},{0,2,1,2}},

  {{0,3,0,0},{1,2,2,2}}, {{0,3,0,1},{1,2,3,2}}, {{0,3,0,2},{1,2,0,2}},
  {{0,3,1,3},{1,2,1,3}}, {{0,3,2,3},{1,2,1,0}}, {{0,3,3,3},{1,2,1,1}},

  {{1,0,1,1},{2,3,3,3}}, {{1,0,1,2},{2,3,0,3}}, {{1,0,1,3},{2,3,1,3}},
  {{1,0,0,0},{2,3,2,2}}, {{1,0,2,0},{2,3,2,0}}, {{1,0,3,0},{2,3,2,1}},

  {{1,1,1,0},{3,3,2,3}}, {{1,1,1,2},{3,3,0,3}}, {{1,1,1,3},{3,3,1,3}},
  {{1,1,0,1},{3,3,3,2}}, {{1,1,2,1},{3,3,3,0}}, {{1,1,3,1},{3,3,3,1}},

  {{1,3,1,0},{1,3,2,3}}, {{1,3,1,1},{1,3,3,3}}, {{1,3,1,2},{1,3,0,3}},

  {{2,0,2,1},{2,0,3,0}}, {{2,0,2,2},{2,0,0,0}}, {{2,0,2,3},{2,0,1,0}},

  {{2,1,2,0},{3,0,2,0}}, {{2,1,2,2},{3,0,0,0}}, {{2,1,2,3},{3,0,1,0}},
  {{2,1,0,1},{3,0,3,2}}, {{2,1,1,1},{3,0,3,3}}, {{2,1,3,1},{3,0,3,1}},

  {{3,1,3,0},{3,1,2,1}}, {{3,1,3,2},{3,1,0,1}}, {{3,1,3,3},{3,1,1,1}}
};

void read_params(int argc, char* argv[], Params* params);
void read_sequence(Params* params, char* seqfile);
void read_tree(Params* params, char* treefile);
Node* read_node(Params* params, FILE* in, int is_root);
void num_changes_from_sequence(Params* params, ShortNucVec* num_changes, char* seq);
void compute_node(Params* params, Node* node);
void print_tree(Params* params);
void print_node(Params* params, Node* node, int is_root);
void compute_parsimony(Params* params, Node* node, Node* parent);
void sum_changes(Params* params, Node* node, Node* parent, Changes* changes);
void count_changes(Params* params, char* ancestor, char* descendant, Changes* changes, double time, int branch_num);
void print_changes(Changes* changes, int branches);
int get_state(int n, double* vector);

int main(int argc, char* argv[]) {
  Params *params;
  Changes *changes;

  (params = (Params*)malloc(sizeof(Params))) || die("Could not allocate memory for params\n");
  read_params(argc, argv, params);

  randomize();
  fprintf(stderr, "computing nodes...\n");
  compute_node(params, params->root);
  fprintf(stderr, "computing parsimony...\n");
  compute_parsimony(params, params->root, NULL);
  fprintf(stderr, "counting changes...\n");
  (changes = (Changes*)malloc(sizeof(Changes))) || die("Could not allocate memory for changes\n");
  (changes->branch = (Branch*)malloc(sizeof(Branch) * params->num_branches)) || die("Could not allocate memory for changes->branch\n");
  sum_changes(params, params->root, NULL, changes);

  print_changes(changes, params->num_branches);

  return 0;
}

void print_changes(Changes* changes, int branches) {
  int i, j, k, l, sum, b;

  /*
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      if(changes->mononuc[i][j])
	printf("%c -> %c = %5.d [%.5f]\n", NAUC[i], NAUC[j], changes->mononuc[i][j], (double)changes->mononuc[i][j] / changes->monoback[i]);
  */
  /*
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++)
	for(l = 0; l < NUM_NUC; l++)
	  if(changes->dinuc[i][j][k][l])
	    printf("%c%c -> %c%c = %5.d [%.5f] (%.5f)\n", NAUC[i], NAUC[j], NAUC[k], NAUC[l], 
		   changes->dinuc[i][j][k][l], (double)changes->dinuc[i][j][k][l] / changes->diback[i][j], 
		   ((double)changes->mononuc[i][k] / changes->monoback[i]) 
		   * ((double)changes->mononuc[j][l] / changes->monoback[j]));
		   */
  /*
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++)
	for(l = 0; l < NUM_NUC; l++)
	  if((i * 64 + j * 16 + k * 4 + l <= complement[j] * 64 + complement[i] * 16 + complement[l] * 4 + complement[k]) && (i == k || j == l) && !(i == k && j == l))
	    printf("%c%c->%c%c = %c%c->%c%c = %.7f %.5f (%.5f)\n", 
		   NAUC[i], NAUC[j], NAUC[k], NAUC[l], 
		   NAUC[complement[j]], NAUC[complement[i]], NAUC[complement[l]], NAUC[complement[k]], 
		   changes->symdinucrate[i][j][k][l] / changes->time, 
		   (double)changes->symdinuc[i][j][k][l] / changes->symdiback[i][j], 
		   (i == k) ? ((double)changes->mononuc[j][l] / changes->monoback[j]) : ((double)changes->mononuc[i][k] / changes->monoback[i]));
  */

  printf("%d:      ", branches);
  for(b = 0; b < branches; b++)
    printf(" %15.10lg", changes->branch[b].length);
  printf("\n");
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++)
	for(l = 0; l < NUM_NUC; l++)
	  if(1 || changes->trinuc[i][j][k][l]) {
	    /*
	    printf("%c%c%c -> %c%c%c = %5.d [%.5f] (%.5f)\n", 
		   NAUC[i], NAUC[j], NAUC[k], NAUC[i], NAUC[l], NAUC[k], 
		   changes->trinuc[i][j][k][l], 
		   (double)changes->trinuc[i][j][k][l] / changes->triback[i][j][k], 
		   ((double)changes->mononuc[i][i] / changes->monoback[i]) 
		   * ((double)changes->mononuc[j][l] / changes->monoback[j]) 
		   * ((double)changes->mononuc[k][k] / changes->monoback[k]));
		   */
	    printf("%c%c%c -> %c%c%c: ", NAUC[i], NAUC[j], NAUC[k], NAUC[i], NAUC[l], NAUC[k]);
	    for(b = 0; b < branches; b++)
	      printf("%8.1d ", changes->branch[b].trinuc[i][j][k][l]);
	    printf("\n");
	  }
  /*
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++) {
	for(sum = l = 0; l < NUM_NUC; l++)
	  sum += changes->trinuc[i][j][l][k];
	printf("%c%c -> %c%c = %5.d (%5.d)\n", NAUC[i], NAUC[j], NAUC[i], NAUC[k], sum, changes->dinuc[i][j][i][k]);
      }
  */
}

void sum_changes(Params* params, Node* node, Node* parent, Changes* changes) {
  int n, i, j, k, l, b;

  if(parent) {
    changes->branch[node->branch_num].length = node->time_to_parent;
    count_changes(params, parent->pars_seq, node->pars_seq, changes, node->time_to_parent, node->branch_num);
  }
  else {
    for(i = 0; i < NUM_NUC; i++) {
      changes->monoback[i] = 0;
      changes->symmonoback[i] = 0;
      for(j = 0; j < NUM_NUC; j++) {
	changes->mononuc[i][j] = 0;
	changes->symmononuc[i][j] = 0;
	changes->diback[i][j] = 0;
	changes->symdiback[i][j] = 0;
	for(k = 0; k < NUM_NUC; k++) {
	  changes->triback[i][j][k] = 0;
	  for(l = 0; l < NUM_NUC; l++) {
	    changes->dinuc[i][j][k][l] = 0;
	    changes->symdinuc[i][j][k][l] = 0;
	    changes->trinuc[i][j][k][l] = 0;
	    changes->symdinucrate[i][j][k][l] = 0.0;
	  }
	}
      }
      for(b = 0; b < params->num_branches; b++)
	for(i = 0; i < NUM_NUC; i++)
	  for(j = 0; j < NUM_NUC; j++)
	    for(k = 0; k < NUM_NUC; k++)
	      for(l = 0; l < NUM_NUC; l++) 
		changes->branch[b].trinuc[i][j][k][l] = 0;
    }
    changes->samples = 0;
    changes->time = 0.0;
  }
  for(n = 0; n < node->n_children; n++)
    sum_changes(params, node->children[n], node, changes);
}

void count_changes(Params* params, char* ancestor, char* descendent, Changes* changes, double time, int branch_num) {
  int i, j, k, l;
  int symdinuc[NUM_NUC][NUM_NUC][NUM_NUC][NUM_NUC];
  int symdinucback[NUM_NUC][NUM_NUC];

  /*
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++) {
      symdinucback[i][j] = 0;
      for(k = 0; k < NUM_NUC; k++)
	for(l = 0; l < NUM_NUC; l++)
	  symdinuc[i][j][k][l] = 0;
    }
  for(i = 0; i < params->seq_length; i++)
    if(params->valid_sequence[i]) {
      changes->mononuc[ancestor[i]][descendent[i]]++;
      changes->monoback[ancestor[i]]++;
    }
  for(i = 0; i < params->seq_length; i++)
    if(params->valid_sequence[i]) {
      if(ancestor[i] * 4 + descendent[i] <= complement[ancestor[i]] * 4 + descendent[i]) {
	changes->symmononuc[ancestor[i]][descendent[i]]++;
	changes->symmonoback[ancestor[i]]++;
      }
      else {
	changes->symmononuc[complement[ancestor[i]]][complement[descendent[i]]]++;
	changes->symmonoback[complement[ancestor[i]]]++;
      }
    }
  for(i = 0; i < params->seq_length - 1; i++)
    if(params->valid_sequence[i] && params->valid_sequence[i+1]) {
      changes->dinuc[ancestor[i]][ancestor[i+1]][descendent[i]][descendent[i+1]]++;
      changes->diback[ancestor[i]][ancestor[i+1]]++;
    }
  for(i = 0; i < params->seq_length - 1; i++)
    if(params->valid_sequence[i] && params->valid_sequence[i+1]
       && (ancestor[i] == descendent[i] || ancestor[i+1] == descendent[i+1])) {
      if(ancestor[i] * 64 + ancestor[i+1] * 16 + descendent[i] * 4 + descendent[i+1] <= complement[ancestor[i+1]] * 64 + complement[ancestor[i]] * 16 + complement[descendent[i+1]] * 4 + complement[descendent[i]]) {
	changes->symdinuc[ancestor[i]][ancestor[i+1]][descendent[i]][descendent[i+1]]++;
	changes->symdiback[ancestor[i]][ancestor[i+1]]++;
	symdinuc[ancestor[i]][ancestor[i+1]][descendent[i]][descendent[i+1]]++;
	symdinucback[ancestor[i]][ancestor[i+1]]++;
      }
      else {
	changes->symdinuc[complement[ancestor[i+1]]][complement[ancestor[i]]][complement[descendent[i+1]]][complement[descendent[i]]]++;
	changes->symdiback[complement[ancestor[i+1]]][complement[ancestor[i]]]++;
	symdinuc[complement[ancestor[i+1]]][complement[ancestor[i]]][complement[descendent[i+1]]][complement[descendent[i]]]++;
	symdinucback[complement[ancestor[i+1]]][complement[ancestor[i]]]++;
      }
    }
  for(i = 0; i < params->seq_length - 2; i++)
    if(params->valid_sequence[i] && params->valid_sequence[i+1] && params->valid_sequence[i+2]
       && ancestor[i] == descendent[i] && ancestor[i+2] == descendent[i+2] 
       && !(0 && NAUC[ancestor[i+1]] == 'C' && NAUC[ancestor[i+2]] == 'G')
       && !(0 && NAUC[ancestor[i]] == 'C' && NAUC[ancestor[i+1]] == 'G')) {
      changes->trinuc[ancestor[i]][ancestor[i+1]][ancestor[i+2]][descendent[i+1]]++;
      changes->triback[ancestor[i]][ancestor[i+1]][ancestor[i+2]]++;
      changes->branch[branch_num].trinuc[ancestor[i]][ancestor[i+1]][ancestor[i+2]][descendent[i+1]]++;
    }
  */
  for(i = 0; i < params->seq_length - 2; i++)
    if(params->valid_sequence[i] && params->valid_sequence[i+1] && params->valid_sequence[i+2]
       /*&& descendent[i+1] < NUM_NUC */)
      changes->branch[branch_num].trinuc[ancestor[i]][ancestor[i+1]][ancestor[i+2]][descendent[i+1]]++;
  /*
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++)
	for(l = 0; l < NUM_NUC; l++)
	  if((i * 64 + j * 16 + k * 4 + l <= complement[j] * 64 + complement[i] * 16 + complement[l] * 4 + complement[k]) && (i == k || j == l) && !(i == k && j == l))
	    changes->symdinucrate[i][j][k][l] += (double)symdinuc[i][j][k][l] / symdinucback[i][j];
  */
  changes->samples++;
  changes->time += time;
}

void compute_parsimony(Params* params, Node* node, Node* parent) {
  int n, i, j, min, argmin[NUM_NUC], argn;
  double argp[NUM_NUC];

  if(!node /* || !node->n_children */)
    return;

  for(i = 0; i < params->seq_length; i++)
    if(params->valid_sequence[i]) {
      if(!parent) {
	for(argn = 0, min = BIG_SHORT, j = 0; j < NUM_NUC; j++)
	  if(node->num_changes[i].el[j] < min) {
	    min = node->num_changes[i].el[j];
	    argmin[0] = j;
	    argn = 1;
	  }
	  else if(node->num_changes[i].el[j] == min)
	    argmin[argn++] = j;
	for(j = 0; j < argn; j++) {
	  argp[j] = j ? argp[j-1] : 0.0;
	  for(n = 0; n < node->n_children; n++)
	    if(argmin[j] != node->children[n]->point[i].el[argmin[j]])
	      argp[j] += node->children[n]->time_to_parent;
	}
	node->pars_seq[i] = argmin[get_state(argn, argp)];
      }
      else
	node->pars_seq[i] = node->point[i].el[parent->pars_seq[i]];
    }

  for(n = 0; n < node->n_children; n++)
    compute_parsimony(params, node->children[n], node);
}

int get_state(int n, double* vector) {
  double r;
  int high, low, test;

  if(n == 1)
    return 0;

  r = random_double(0.0, vector[n-1]);

  low = 0;
  high = n - 1;
  while (1) {
    if(low == high)
      return low;
    test = (high + low) / 2;
    if(r < vector[test])
      high = test;
    else
      low = test + 1;
  }
}

void compute_node(Params* params, Node* node) {
  int n, i, j, k, min, argmin[NUM_NUC], argn, changes;

  if(!node || !node->n_children)
    return;

  for(n = 0; n < node->n_children; n++)
    compute_node(params, node->children[n]);
  
  for(i = 0; i < params->seq_length; i++)
    if(params->valid_sequence[i]) {
      for(j = 0; j < NUM_NUC; j++)
	for(node->num_changes[i].el[j] = 0, n = 0; n < node->n_children; n++) {
	  for(argn = 0, min = BIG_SHORT, k = 0; k < NUM_NUC; k++) {
	    changes = node->children[n]->num_changes[i].el[k] + (j == k ? 0 : 1);
	    if(changes < min) {
	      min = changes;
	      argmin[0] = k;
	      argn = 1;
	    }
	    else if(changes == min)
	      argmin[argn++] = k;
	  }
	  node->num_changes[i].el[j] += min;
	  node->children[n]->point[i].el[j] = argmin[random_int(0, argn - 1)];
	}
    }
}

void read_params(int argc, char* argv[], Params* params) {
  int i;

  for(i = 1; i < argc && argv[i][0] == '-'; i++) {

    die(USAGE);
  }
  argc == i + 2 || die(USAGE);
  read_sequence(params, argv[i++]);

  read_tree(params, argv[i]);
}

void read_sequence(Params* params, char* seqfile) {
  FILE* in;
  int i, j, ch;
  char* cp;

  (in = fopen(seqfile, "r")) || die("Could not open seqfile: %s\n", seqfile);
  fscanf(in, "%d%d\n", &params->n_seq, &params->seq_length) == 2 || die("Error in seqfile: %s\n", seqfile);
  (params->sequence = (Sequence*)malloc(sizeof(Sequence) * params->n_seq)) || die("Could not allocate memory for params->sequence\n");
  (params->valid_sequence = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->valid_sequence\n");
  for(i = 0; i < params->seq_length; i++)
    params->valid_sequence[i] = 0;
  for(i = 0; i < params->n_seq; i++) {
    fscanf(in, "%s.63", params->sequence[i].name) == 1 || die("Error in seqfile\n");
    (params->sequence[i].seq = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->sequence[%d].seq\n", i);
    for(j = 0; j < params->seq_length;) {
      if((ch = getc(in)) != EOF && (cp = strchr(NAUC, ch))) {
	if((unsigned int)cp - (unsigned int)NAUC < NUM_NUC)
	  params->valid_sequence[j] = 1;
	params->sequence[i].seq[j++] = (unsigned int)cp - (unsigned int)NAUC;
      }
      else if(ch == '\n' || ch == EOF)
	die("Early sequence termination in sequence: %s, %d expected, %d found\n", params->sequence[i].name, params->seq_length, j);
    }
    for(; ch != '\n' && ch != EOF; ch = getc(in));
  }

  fclose(in);
}

void read_tree(Params* params, char* treefile) {
  FILE* in;

  (in = fopen(treefile, "r")) || die("Could not open treefile: %s\n", treefile);
  params->num_branches = 0;
  params->root = read_node(params, in, 1);
  fclose(in);
}

Node* read_node(Params* params, FILE* in, int is_root) {
  Node* node;
  int i, ch;
  char name[MAX_NAME];

  (node = (Node*)malloc(sizeof(Node))) || die("Could not allocate memory for node\n");
  (node->num_changes = (ShortNucVec*)malloc(sizeof(ShortNucVec) * params->seq_length)) || die("Could not allocate memory for node->num_changes\n");
  node->n_children = 0;
  (node->pars_seq = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for node->pars_seq\n");
  (node->point = (NucVec*)malloc(sizeof(NucVec) * params->seq_length)) || die("Could not allocate memory for node->point\n");


  while(isspace(ch = getc(in)));
  if(ch == '(') {
    for(; ; node->n_children++) {
      while(isspace(ch = getc(in)));
      if(ch == ')')
	break;
      if(ch != ',')
	ungetc(ch, in);
      node->n_children < MAX_CHILDREN || die("Node has too many children\n");
      !node->n_children || ch == ',' || die("Missing comma in treefile\n");
      node->children[node->n_children] = read_node(params, in, 0);
      node->seq = -1;
    }
  }
  else {
    ungetc(ch, in);
    fscanf(in, "%[^\n:].63", name) == 1 || die("Missing name in treefile\n");
    for(i = 0; i < params->n_seq; i++)
      if(!strcmp(params->sequence[i].name, name)) {
	num_changes_from_sequence(params, node->num_changes, params->sequence[i].seq);
	memcpy(node->pars_seq, params->sequence[i].seq, sizeof(char) * params->seq_length);
	break;
      }
    if(i == params->n_seq)
      die("Name %s in tree file but not found in sequence file\n", name);
    node->seq = i;
  }
  if(!is_root) {
    while(isspace(ch = getc(in)));
    ch == ':' || die("Missing colon in treefile after %s\n", name);
    while(isspace(ch = getc(in)));
    ungetc(ch, in);
    fscanf(in, "%lg", &node->time_to_parent) == 1 || die("Missing time to parent\n"); 
    node->branch_num = params->num_branches++;
  }
  else {
    while(isspace(ch = getc(in)));
    ch == ';' || die("Missing semicolon at end of treefile\n");
  }
  return node;
}

void num_changes_from_sequence(Params* params, ShortNucVec* num_changes, char* seq) {
  int i, j;

  for(i = 0; i < params->seq_length; i++)
    if(params->valid_sequence[i])
      for(j = 0; j < NUM_NUC; j++) {
	if(j == seq[i])
	  num_changes[i].el[j] = 0;
	else if(seq[i] < NUM_NUC)
	  num_changes[i].el[j] = BIG_SHORT;
	else
	  num_changes[i].el[j] = 0;
      }
}

void print_tree(Params* params) {
  if(params->root)
    print_node(params, params->root, 1);
}

void print_node(Params* params, Node* node, int is_root) {
  int n;

  if(node->n_children > 0) {
    printf("(");
    for(n = 0; n < node->n_children; n++) {
      if(n)
	printf(",");
      print_node(params, node->children[n], 0);
    }
    printf(")");
  }
  else
    printf(params->sequence[node->seq].name);
  if(!is_root)
    printf(":%g", node->time_to_parent);
  else
    printf(";\n");
}



